from typing import Type
import os
from pathlib import Path
from .controller import Controller
from .store import UserStore
from .autoloader import AutoLoader
from inspect import isclass
from .config import load_config
from .model import Model


class Loader:
    autoloader: AutoLoader

    def __init__(self):
        self.autoloader = AutoLoader()

    def load_controllers(self, path: str) -> list[Type[Controller]]:
        _objects = self.autoloader.load_objects(path, recursive=True)
        return list(filter(lambda obj: issubclass(obj.__class__, Controller), _objects))

    def load_middlewares(self):
        pass

    def load_user_stores(self, path: str) -> list[Type[UserStore]]:
        _objects = self.autoloader.load_objects(path, recursive=True)
        return list(filter(lambda obj: isclass(obj) and issubclass(obj, UserStore) and obj != UserStore, _objects))

    def load_models(self, path: str):
        _objects = self.autoloader.load_objects(path, recursive=True)
        return list(filter(lambda obj: isclass(obj) and issubclass(obj, Model) and obj != Model, _objects))

    def load_config(self, path: str, config_class_name: str, dot_env_file_name: str = '.env'):
        EnvConfig = self.autoloader.load_class(path, config_class_name)
        working_dir = os.getcwd()
        path_to_config_file = Path(working_dir).joinpath(dot_env_file_name)
        print(path_to_config_file)
        config = load_config(EnvConfig, str(path_to_config_file))
        return config





from aioschedule import run_pending
from asyncio import sleep
from typing import NoReturn


async def start_scheduling() -> NoReturn:
    """
    Start async scheduler forever
    """
    while True:
        await run_pending()
        await sleep(1)



from typing import Callable
from telgine.utils.make_async import wrap_async
from abc import ABCMeta, abstractmethod
from types import CoroutineType
from typing import Type
from .telgine_object import TelgineObject
from aiogram import Dispatcher


class Controller(TelgineObject):
    __metaclass__ = ABCMeta
    handler: CoroutineType

    def __init__(self, handler: Callable | CoroutineType, *args, **kwargs):
        self.handler = wrap_async(handler)


def controller(class_: Type[Controller], require_parentheses: bool | None | Callable = None):
    def decorate(fn: Callable[[...], any]):
        is_parentheses_exist: bool

        def dec(*args, **kwargs):
            nonlocal is_parentheses_exist
            is_parentheses_exist = not (len(args) == 1 and callable(args[0]))

            def wrap(func_: Callable):
                controller_object = class_(func_, *args, **kwargs) if is_parentheses_exist else class_(func_, *(args[1:]))
                return controller_object

            if require_parentheses is None:
                if is_parentheses_exist:
                    return wrap
                return wrap(args[0])
            elif require_parentheses:
                if is_parentheses_exist:
                    return wrap
                raise RuntimeError('Parentheses are required, but they are not')
            elif not require_parentheses:
                if not is_parentheses_exist:
                    return wrap(args[0])
                raise RuntimeError('Parentheses are not required, but they are')

        return dec

    if callable(require_parentheses):
        # If decorator using without parentheses
        func = require_parentheses
        require_parentheses = None
        return decorate(func)
    return decorate



class EnvConfig:
    BOT_NAME: str
    BOT_TOKEN: str
    ADMINS: list[int]

    APP_HOST: str = 'localhost'
    APP_PORT: int = 8443

    DB_TYPE: str
    DB_HOST: str
    DB_PORT: int
    DB_NAME: str
    DB_USER: str
    DB_PASSWORD: str
    DB_PATH: str


config = EnvConfig



#  This is Entry point to all framework. In this file going on
from telgine.utils.singleton import singleton
from .register import Register
from .bot import Bot
from aiogram import Dispatcher
from .store import MemoryStorage
from peewee import Database
from .db import connect_db, DBType


@singleton
class Telgine:
    """
    Application entry point.
    """
    bot: Bot
    register: Register
    db: Database
    config: any

    def __init__(self):
        self.register = Register()
        self.config = self.register.register_config('bot/config/env.py')

        self.bot = Bot(self.config.BOT_TOKEN)

        self.register.register_user_stores('bot/stores', MemoryStorage())
        self.register.register_controllers('bot/controllers', self.bot.dp)
        self.register.register_middlewares(self.bot.dp)
        self.bot.startup_callback = self.register.on_startup
        self.db = connect_db(
            DBType(self.config.DB_TYPE),
            self.config.DB_PATH,
            self.config.DB_NAME,
            self.config.DB_USER,
            self.config.DB_PASSWORD,
            self.config.DB_HOST,
            self.config.DB_PORT
        )
        models = self.register.register_models('bot/models', self.db)


    def run(self, skip_updates: bool = False):
        self.bot.startup(skip_updates)


def get_app() -> Telgine:
    """
    Return main instance of application. If it's not there then create and return it
    :return: application instance
    """
    return Telgine()


def get_dp() -> Dispatcher:
    """
    Return aiogram bot dispatcher
    :return: aiogram bot dispatcher
    """
    return get_app().bot.dp


def get_bot() -> Bot:
    """
    Return instance of class Bot
    :return: application config
    """
    return get_app().bot


def get_config():
    """
    Return application config
    :return: application config
    """
    return get_app().config



from abc import abstractmethod, ABCMeta


class TelgineObject:
    __meta_class_ = ABCMeta

    @abstractmethod
    def on_register(self, *args, **kwargs):
        """Calling when an object will be registered"""



from .app import get_app as app, get_dp as dp, get_bot as bot, get_config as config, Telgine
from .ctx import message
from .functions import send, ask_chat, ask
from .controllers import command, startup, halt, every, SECOND
from .store import Storage, MemoryStorage, UserStore
from .config import load_config
from .model import Model, RelationField, EnumField, inverse_relations



from typing import Callable, Literal
from aiogram.types.reply_keyboard import ReplyKeyboardMarkup, KeyboardButton
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import Element
from abc import ABC
from typing import ClassVar, Type


class BaseWidget(ABC):
    pass


class ChildrenWidget(BaseWidget):
    children: list[BaseWidget] = []

    def __init__(self, *args):
        self.children = list(args)

    def add(self, children: BaseWidget):
        self.children.append(children)


class If(ChildrenWidget):
    test: str
    orelse: list[BaseWidget]


class View(ChildrenWidget):
    width: int
    direction: Literal['row', 'column']

    def __init__(self, *args):
        super().__init__(*args)


class Keyboard(View):
    children: list = []

    def __init__(self, *args, width: int | None = None):
        super().__init__(*args)
        self.children = list(args)
        self.width = width


class Button(BaseWidget):
    title: str
    on_click: Callable

    def __init__(self, title: str, on_click: Callable | str | None = None):
        self.title = title
        self.on_click = on_click


Keyboard(
    Button('One', lambda: print('Hello World')),
    Button('Two', lambda: print('Hello World')),
    Button('Two', lambda: print('Hello World')),
    Button('Two', lambda: print('Hello World')),
    Button('Two', lambda: print('Hello World')),
    Keyboard(
        Button('Another', lambda: print('Another')),
        Button('Another', lambda: print('Another')),
        Button('Another', lambda: print('Another')),
        Button('Another', lambda: print('Another')),
    ),
    width=5,
)





# Database initialization for all application
from peewee import Database, SqliteDatabase, MySQLDatabase, PostgresqlDatabase, Model
from typing import Type
from enum import Enum


class DBType(Enum):
    POSTGRES = 'POSTGRES'
    SQLITE = 'SQLITE'
    MYSQL = 'MYSQL'


def connect_db(
        db_type: DBType,
        path: str | None,
        database: str | None,
        user: str | None,
        password: str | None,
        host: str | None,
        port: int | None
) -> Database:
    if db_type == DBType.SQLITE:
        return SqliteDatabase(path)
    elif db_type == DBType.POSTGRES:
        return PostgresqlDatabase(database=database, user=user, password=password, host=host, port=port)
    elif db_type == DBType.MYSQL:
        return MySQLDatabase(database=database, user=user, password=password, host=host, port=port)
    raise TypeError('Unsupported DB type')


def inject_db(db: Database, model: Type[Model]):
    getattr(model, '_meta').database = db
    return model



from typing import Callable
from .is_async import is_async
from types import CoroutineType


def wrap_async(func: Callable | CoroutineType):
    async def wrapper(*args, **kwargs):
        if is_async(func):
            await func()
        else:
            func()
    return wrapper



import abc
from abc import ABCMeta
from .ctx import message
from telgine.utils.get_class_fields import get_class_fields, ClassField

class Storage:
    """
    Abstract object storage for storing data
    """
    __metaclass__ = ABCMeta

    @abc.abstractmethod
    def __setitem__(self, key, value):
        pass

    @abc.abstractmethod
    def __getitem__(self, item):
        pass

    @abc.abstractmethod
    def __delitem__(self, item):
        pass


class MemoryStorage(Storage):
    """
    Specific implementation of Storage. Store all data in memory.
    """
    storage = {}
    current = 0

    def __setitem__(self, key, value):
        self.storage[key] = value

    def __getitem__(self, item):
        return self.storage[item]

    def __delitem__(self, item):
        del self.storage[item]

    def __iter__(self):
        return self

    def __next__(self):
        if self.current + 1 > len(self.storage):
            self.current = 0
            raise StopIteration
        key = list(self.storage)[self.current]
        val = self.storage[key]
        self.current += 1
        return val


class UserStore:
    """
    Class for creating Stores for temporary storing local user data outside of request handlers.
    """
    #  TODO; Solve problem with setters, getters and deleters for attributes of UserStore child classes.
    #   Probably I need to patch AST ("class.attr = val" replace to "class.set('attr', val)")
    _initialized = False
    _first_getting_attr: bool = True
    _storage: Storage
    _fields: list[ClassField] = []
    _instance = None

    def __init__(self):
        if self._initialized:
            return
        self._fields = get_class_fields(self)
        for field in self._fields:
            setattr(self, field.name, None)
        self._initialized = True

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(UserStore, cls).__new__(cls)
        return cls._instance

    def __setattr__(self, key, value):
        if not self._initialized or key.startswith('_'):
            self.__dict__[key] = value
            return
        # if user data object doesn't exist in storage then create and initialize it
        chat_id = message().chat.id
        if chat_id not in self._storage:
            self._storage[chat_id] = {}
            self._reset()
        self._storage[chat_id][key] = value
        self.__dict__[key] = value

    def __getattr__(self, item):
        if item.startswith('_'):
            return self.__dict__[item]
        if self._first_getting_attr:
            self._reset()
            self._first_getting_attr = False
        chat_id = message().chat.id
        return self._storage[chat_id][item]

    def __delattr__(self, item):
        if item.startswith('_'):
            del self.__dict__[item]
            return
        chat_id = message().chat.id
        if item in self._storage[chat_id]:
            del self._storage[chat_id][item]
            delattr(self, item)

    def clear(self):
        """Delete all data from storage for this object"""
        chat_id = message().chat.id
        for field in self._fields:
            del self._storage[chat_id][field.name]

    def _reset(self):
        """
        Reset all data from storage for this object to default state and values
        :return:
        """
        for field in self._fields:
            if field.value is not None:
                setattr(self, field.name, field.value)
            else:
                delattr(self, field.name)

    def delete(self):
        """
        Delete user data object from store
        :return:
        """
        chat_id = message().chat.id
        del self._storage[chat_id]


# This is autoloader which load all classes, objects, handlers from the files

import inspect
from types import ModuleType, FunctionType
from typing import Callable
import glob
import os
from pathlib import Path
import importlib.util
import importlib.machinery
from ..utils.prefix import remove_prefix


def load_module(base_path, module_path):
    # FIXME: fix path converting
    module_path = str(module_path)
    if module_path[0] == '/':
        module_path = module_path[1:]
    python_module_path = module_path.split('.')[0].replace('/', '.')
    package = '.'.join(python_module_path.split('.')[:-1])
    module = importlib.import_module(python_module_path, package)
    return module


class AutoLoader:
    base_path: Path

    def __init__(self, base_path=os.getcwd()):
        self.base_path = Path(base_path)

    def load_modules(self, path, recursive: bool = False) -> ModuleType | list[ModuleType]:
        """
        Load all modules by path. If path end to directory then load and return all modules in this directory.
        If path end to file then load and return only module of this file.

        :param path: path to directory of modules or module
        :param recursive: load modules recursively for directories
        """
        path = Path(path)

        if path.is_absolute():
            result_path = path
        else:
            result_path = self.base_path.joinpath(path)

        #  If path is path to file
        if path.is_file():
            module = load_module(self.base_path, path)
            return module
            #  TODO: Prevent several modules reloadings (probably problem with module names)
            #   By this reason doesn't work Config and UserStore
            #   Sequence of module loading so very important (app.py)

        #  If path is path to directory then load all modules in this directory
        modules = []
        find_path = str(result_path.joinpath('**/*.py'))
        for filename in glob.iglob(find_path, recursive=recursive):
            # FIXME: fix path converting
            module = load_module(self.base_path, Path(remove_prefix(str(self.base_path), filename)))
            modules.append(module)
        return modules

    def _module_object_type_filter(self, filter_: Callable, path: str, recursive: bool = False):
        def object_filter(_filter, module):
            objects_names = dir(module)
            objects_ = list(map(lambda item: getattr(module, item), objects_names))
            filtered_objects = list(filter(lambda item: _filter(item), objects_))
            return filtered_objects

        modules = self.load_modules(path, recursive=recursive)
        objects = []
        if type(modules) == list:
            for modul in modules:
                objects += object_filter(filter_, modul)
            return objects

        objects += object_filter(filter_, modules)
        return objects

    def load_functions(self, path: str, recursive=False):
        """
        Will load and return all functions by path.
        :param path:
        :param recursive:
        :return:
        """
        return self._module_object_type_filter(lambda obj: type(obj) == FunctionType, path, recursive)

    def load_objects(self, path: str, type_: any = None, recursive=False):
        """
        Will load and return all objects
        :param path:
        :param type_:
        :param recursive:
        :return:
        """
        if type_ is None:
            return self._module_object_type_filter(lambda obj: obj, path, recursive)
        return list(set(self._module_object_type_filter(lambda obj: type(obj) == type_, path, recursive)))

    def load_function(self, path: str, function_name, recursive=False):
        """
        Will load and return only one function the name of which set in attributes.
        :param path:
        :param function_name:
        :param recursive:
        :return:
        """
        result = list(
            filter(lambda item: item.__name__ == function_name, self.load_functions(path, recursive=recursive))
        )
        if len(result) == 0:
            raise Exception(f'Can not find name {function_name}')
        return result[0]

    def load_classes(self, path, recursive=False):
        """
        Will load and return all classes by path.
        :param path:
        :param recursive:
        :return:
        """
        return self._module_object_type_filter(lambda obj: inspect.isclass(obj), path, recursive)

    def load_class(self, path, class_name, recursive=False):
        """
        Will load and return only one class the name of which set in attributes.
        :param path:
        :param class_name:
        :param recursive:
        :return:
        """
        result = list(filter(lambda item: item.__name__ == class_name, self.load_classes(path, recursive)))
        if len(result) == 0:
            raise Exception(f'Can not find name {class_name}')
        return result[0]



from sly import Lexer


class TelgineLexer(Lexer):
    tokens = {ID, GT, LT, CLOSE, ASSIGN, TEXT, STRING, COLON, IF, ELSE, FOR}

    @_(r'(?<=>\n)([^<|>|])+(?=<)|(?<=>)([^<|>|])+(?=<)')
    def TEXT(self, t):
        t.value = t.value.strip()

        if len(t.value) > 0:
            return t

    CLOSE = r'</'
    GT = r'>'
    LT = r'<'
    ASSIGN = r'='

    STRING = r'"([^"]+)?"'

    ID = r'[A-z]([A-z]|\d||\-)+[A-z\d]'
    ID['v-if'] = IF
    ID['v-else'] = ELSE
    ID['v-for'] = FOR

    @_(r'"(([^"])+)"')
    def STRING(self, t):
        t.value = t.value[1:-1]
        return t
    COLON = r':'

    ignore = r'\f\r\t\n '
    ignore_newline = r'\n+'


def tokenize(string):
    lexer = TelgineLexer()
    return lexer.tokenize(string)



from aiogram import Dispatcher, Bot as AiogramBot, executor
from aiogram.contrib.fsm_storage.memory import MemoryStorage
from types import CoroutineType


class Bot:
    dp: Dispatcher
    bot: AiogramBot
    startup_callback: CoroutineType

    def __init__(self, bot_token: str):
        STORAGE = MemoryStorage()
        self.bot = AiogramBot(bot_token)
        self.dp = Dispatcher(self.bot, STORAGE)

    def startup(self, skip_updates):
        #  FIXME: fix problem with typing
        async def on_startup(_):
            await self.startup_callback()
        executor.start_polling(self.dp, skip_updates=skip_updates, on_startup=on_startup)



from typing import Optional
from dataclasses import dataclass


def cast_dotenv_string_to_list(string: str, list_type: any):
    string = ''.join(filter(lambda char: char != ' ', string))
    items = string.split(',')
    items = list(map(lambda item: list_type(item), items))
    return list(items)


@dataclass
class ClassField:
    name: str
    type: any
    value: Optional[any]


def get_class_fields(_class, only_public=False) -> list[ClassField]:
    class_fields = {}

    # for type annotated fields

    fields_with_value = list(
        filter(lambda k: not ((k.startswith('__') and k.endswith('__')) or k.startswith('_')), _class.__dict__.keys())
    )

    for key in fields_with_value:
        value = getattr(_class, key)
        _type = type(value)
        class_fields[key] = ClassField(key, _type, value)

    for key, _type in _class.__annotations__.items():
        if key not in class_fields:
            class_fields[key] = ClassField(key, _type, None)

    return list(class_fields.values())



from aiogram.dispatcher.middlewares import BaseMiddleware
from aiogram.types import Message
from werkzeug.local import LocalStack, LocalProxy, release_local

#  Context object for storing request data
_request = LocalStack()


#  TODO: Test working context with different users and fix it
def message() -> Message:
    """
    Return request message from context (like request() in Flask)
    :return: request message
    """
    if not hasattr(_request, 'message'):
        raise RuntimeError('message function must used in request context')
    return getattr(_request, 'message')


class OpenContextMiddleware(BaseMiddleware):
    """
    Middleware for creating request context per each request.
    """
    async def on_process_message(self, msg: Message, _):
        _request.message = msg


class CloseContextMiddleware(BaseMiddleware):
    """
    Middleware for destroying request context after all handlers and middlewares.
    """
    async def on_process_message(self, _, __):
        release_local(_request)


class State:
    type: any
    storage = {}

    def __init__(self):
        pass

    def __setattr__(self, key, value):
        msg = message()
        # if storage has not instanced in storage for user then create it
        if not hasattr(self.storage, msg.chat.id):
            self.storage[msg.chat.id] = self.type()
        setattr(self.storage[msg.chat.id], key, value)

    def __delattr__(self, item):
        del self.storage[item][item]

    def __getattr__(self, item):
        msg = message()
        return self.storage[msg.chat.id][item]



# This file contain several controller decorators such as:
# - command - add handler on command
# - startup - add handler on application startup
# - halt - add ahndler on application halt
# - every - call something function every time
# -

from types import CoroutineType
from typing import Callable
from enum import Enum
from functools import wraps

import aioschedule as schedule
from aiogram import Dispatcher

from telgine.utils.is_async import is_async
from .controller import controller, Controller
from telgine.utils.prefix import remove_postfix, remove_prefix


class Command(Controller):
    name: str
    description: str | None

    def __init__(self, handler: Callable | CoroutineType, name: str | None = None, description: str | None = None):
        super().__init__(handler)
        handler.__name__ = remove_postfix('(_+|_command)', handler.__name__)
        handler.__name__ = remove_prefix('_+', handler.__name__)
        self.name = handler.__name__ if name is None else name
        self.description = description


@controller(class_=Command)
def command(name: str | None = None, description: str = 'placeholder description'): pass


class OnStartup(Controller): pass


@controller(class_=OnStartup)
def startup(): pass


class OnHalt(Controller): pass


@controller(class_=OnHalt)
def halt(): pass


# WeekDay definition
class TimeType(Enum):
    DAY = 0


SECOND = (TimeType.DAY, -1)

DAY = (TimeType.DAY, 0)
MONDAY = (TimeType.DAY, 1)
TUESDAY = (TimeType.DAY, 2)
WEDNESDAY = (TimeType.DAY, 3)
THURSDAY = (TimeType.DAY, 4)
FRIDAY = (TimeType.DAY, 5)
SATURDAY = (TimeType.DAY, 6)
SUNDAY = (TimeType.DAY, 7)


# Usage
# @every(MONDAY, TUESDAY, at='10:20')
# def send_notification():
#   pass
def every(*kwargs, at: str | None = None):
    def decorate(func: Callable):
        @wraps(func)
        def wrapper():
            day_map = {
                SECOND: 'second',
                DAY: 'day',
                MONDAY: 'monday',
                TUESDAY: 'tuesday',
                WEDNESDAY: 'wednesday',
                THURSDAY: 'thursday',
                FRIDAY: 'friday',
                SATURDAY: 'saturday',
                SUNDAY: 'sunday'
            }
            listed_args = list(kwargs)
            days: list = list(filter(lambda item: item[0] == TimeType.DAY, listed_args))
            time = at
            prepare_job = schedule.every()
            for day in days:
                day = day_map[day]
                prepare_job = getattr(prepare_job, day)
            if time is not None and time != '':
                if len(days) == 0:
                    prepare_job = prepare_job.day.at(time)
                else:
                    prepare_job = prepare_job.at(time)

            async def async_func():
                if is_async(func):
                    return await func()
                func()

            job = prepare_job.do(async_func)
            return job

        return wrapper

    return decorate



from telgine.core import command, send, config, startup, ask
from bot.stores import User

@startup
async def startup():
    for admin in config().ADMINS:
        await send('Hello, bot started', chat=admin)


@command
async def start():
    user = User()
    await send(f'Bot started')
    user.name = (await ask('Enter your name')).text
    user.age = int((await ask('Enter your name')).text)

@command
async def info():
    user = User()
    if user.name is None or user.age is None:
        return await send('Please set your data')
    await send(f'Hello {user.name} with {user.age} years old!')




from typing import Type
from .controllers import Controller, Command, OnStartup
from .store import UserStore, Storage
from .loader import Loader
from .ctx import OpenContextMiddleware, CloseContextMiddleware
from .ask_data_middleware import AskDataMiddleware
from .model import inverse_relations
from .db import inject_db
from peewee import Database


class Register:
    """
    Class for registering all needless objects for app.
    """
    user_stores: list[UserStore] = []
    controllers: list[Type[Controller]] = []
    loader: Loader

    ask_data_middleware: AskDataMiddleware

    def __init__(self):
        self.loader = Loader()

    def register_controllers(self, path: str, dispatcher):
        self.controllers = self.loader.load_controllers(path)
        for controller in self.controllers:
            if controller.__class__ == Command:
                controller: Command
                dispatcher.register_message_handler(controller.handler, commands=[controller.name])

    async def on_startup(self):
        for controller in self.controllers:
            if controller.__class__ == OnStartup:
                controller: OnStartup
                await controller.handler()

    def register_middlewares(self, dispatcher):
        self.ask_data_middleware = AskDataMiddleware(dispatcher)
        MIDDLEWARES = [
            OpenContextMiddleware(),
            self.ask_data_middleware,
            CloseContextMiddleware()
        ]
        for middleware in MIDDLEWARES:
            dispatcher.setup_middleware(middleware)

    def register_user_stores(self, path: str, storage: Storage):
        user_stores = self.loader.load_user_stores(path)
        for user_store in user_stores:
            store = user_store()
            # FIXME: fix private variable assignment
            store._storage = storage
            self.user_stores.append(store)

    def register_models(self, path: str, db: Database):
        models = self.loader.load_models(path)
        inverse_relations(models)
        for model in models:
            inject_db(db, model)
        return models

    def register_config(self, path: str):
        return self.loader.load_config(path, 'EnvConfig', '.env')



from .core import config, bot, app, dp, Telgine,\
    message, send, ask, ask_chat, command, every,\
    UserStore, Model, EnumField, RelationField,\
    MemoryStorage, startup, inverse_relations,\
    load_config, halt, Storage
from peewee import *

__version__ = '0.2.1'



class Node:
    body: list['Node']
    parent: 'Node'

    def add(self, node):
        node.parent = self
        self.body.append(node)

    def next(self) -> 'Node' | None:
        try:
            next_node_index = self.parent.body.index(self) + 1
        except IndexError:
            return None
        return self.parent.body[next_node_index]

    def prev(self) -> 'Node' | None:
        try:
            prev_node_index = self.parent.body.index(self) - 1
        except IndexError:
            return None
        return self.parent.body[prev_node_index]


class Expr(Node):
    def __init__(self, expr):
        self.expr = expr


class If(Node):
    test: Expr
    orelse: list[Node]

    def __init__(self, test, body: list[Node] | None = None, orelse: list[Node] | None = None):
        if orelse is None:
            orelse = []
        if body is None:
            body = []
        self.test = test
        self.body = body
        self.orelse = orelse

    def orelse_add(self, node: Node):
        node.parent = self
        self.orelse.append(node)


class For(Node):
    test: Expr

    def __init__(self, test: Expr, body: list[Node] | None = None):
        self.test = test
        self.body = body


class Else: pass


class Text(Node):
    text: str

    def __init__(self, text: str):
        self.text = text

    def __repr__(self):
        return self.text


class Tag(Node):
    name: str
    attrs: dict[str, str]

    def __init__(self, name: str, body: list[Node] | None = None, attrs: dict[str, str] | None = None):
        self.name = name
        self.body = body
        self.attrs = attrs

    def __repr__(self):
        return f'<{self.name}/>'

    def find(self, name) -> 'Tag' | None:
        for node in self.body:
            if node.name == name:
                return node
            return self.find(name)
        return None

    def mount(self):
        pass

    def unmount(self):
        pass

    def update(self):
        pass

    def render(self):
        pass


class Button(Tag):
    def render(self):
        pass



class Group(Tag):
    pass


class Message(Tag):
    pass


class Keyboard(Tag):
    pass



from aiogram.types.reply_keyboard import ReplyKeyboardMarkup, ReplyKeyboardRemove
from aiogram.types.inline_keyboard import InlineKeyboardMarkup
from .nodes import Tag



def create_component(children, props) -> Component:
    return Component(children, props)


def render(root: Tag) -> ReplyKeyboardMarkup | ReplyKeyboardRemove | InlineKeyboardMarkup:
    tree = None

    message = root.find('Message')
    if message is not None:
        keyboard = message.find('KeyBoard')
        text = message.find('Text')



    raise NameError(f"Root node can be only <Message/> or <KeyBoard/> components")




from ..core.register import Register
from ..core.db import connect_db, DBType
import pathlib
import os


def migrate_db():
    register = Register()
    workdir = os.getcwd()
    config = register.register_config(str(pathlib.Path(workdir).joinpath('bot/config/env.py')))
    db = connect_db(
        DBType(config.DB_TYPE),
        config.DB_PATH,
        config.DB_NAME,
        config.DB_USER,
        config.DB_PASSWORD,
        config.DB_HOST,
        config.DB_PORT
    )
    models = register.register_models(str(pathlib.Path(workdir).joinpath('bot/models')), db)
    db.create_tables(models)


from typing import TypeVar

T = TypeVar('T')


def singleton(class_: T) -> T:
    INSTANCE_ATTRIBUTE_NAME = '__instance__'

    def getinstance(*args, **kwargs):
        if not hasattr(class_, INSTANCE_ATTRIBUTE_NAME) or type(getattr(class_, INSTANCE_ATTRIBUTE_NAME)) != class_:
            setattr(class_, INSTANCE_ATTRIBUTE_NAME, class_(*args, **kwargs))
        return getattr(class_, INSTANCE_ATTRIBUTE_NAME)
    return getinstance



import click
from .init import initialize_project
from .generate import generate_artifacts
from .migrate import migrate_db
import os
import pathlib

@click.group()
def cli():
    pass


@cli.command()
@click.argument('project_path', type=click.Path())
def init(project_path):
    path = pathlib.Path(project_path)
    if not path.is_absolute():
        path = pathlib.Path(os.getcwd()).joinpath(project_path)
    initialize_project(path)


@cli.command()
@click.argument('object_type')
@click.argument('generated_file')
def create(object_type: str, generated_file):
    generate_artifacts(object_type, generated_file)


@cli.command()
def migrate():
    migrate_db()


def main():
    cli()



# Something useful utilities and function for building chat bot

from ..utils import fmt
from aiogram import types
import asyncio

from .app import get_bot, get_app
from .ctx import message


#  TODO: Make arguments more simple for answer, ask_chat, ask functions, remove doesn't needless arguments
async def _send(
        text: str,
        web_preview: bool = True,
        notify: bool = True,
        protect: bool = False,
        reply_: bool = False,
        chat: int | str | None = None,
) -> types.Message:
    PARSE_MODE = 'Markdown'
    """
    Send the message with text
    :param text: message text
    :param web_preview: web preview for links
    :param notify: send the message silently
    :param protect: protect content for copping and sharing
    :param chat: `None` if your want to answer for a previous user message else user id or @username
    :return:
    """
    if chat is not None:
        return await get_bot().bot.send_message(
            chat_id=chat,
            text=fmt.multiline(text),
            disable_web_page_preview=None if web_preview is None else not web_preview,
            disable_notification=None if notify is None else not notify,
            protect_content=protect,
            parse_mode=PARSE_MODE
        )
    return await message().answer(
        text=fmt.multiline(text),
        disable_web_page_preview=web_preview,
        disable_notification=notify,
        protect_content=protect,
        reply=reply_,
        parse_mode=PARSE_MODE
    )
#  TODO: rename anwere to sendtext


async def send(
        text: str,
        web_preview: bool = True,
        notify: bool = True,
        protect: bool = False,
        chat: int | str | None = None,
):
    return await _send(
        text,
        web_preview,
        notify,
        protect,
        False,
        chat,
    )


async def reply(
    text: str,
    web_preview: bool = True,
    notify: bool = True,
    protect: bool = False,
    chat: int | str | None = None
):
    """
    Send reply message for previously message.
    :return:
    """
    return await _send(
        text,
        web_preview,
        notify,
        protect,
        True,
        chat,
    )


async def ask_chat(chat_id, text) -> types.Message:
    await send(text, chat=chat_id)

    future: asyncio.Future[types.Message] = asyncio.get_event_loop().create_future()

    app = get_app()
    app.register.ask_data_middleware.subscribe(chat_id, future)

    return await future


async def ask(text, chat: int | str | None = None) -> types.Message:
    """
    Send user a message with text argument and await for response from user.
    :param text: asking text
    :param chat: chat to sending. If None then send to chat from context
    :return: response message from user
    """
    # Fixme: need to stop event propagation message in AskDataMiddleware
    if chat is not None:
        return await ask(text, chat)
    msg = message()
    return await ask_chat(msg.chat.id, text)



import re
from .prefix import remove_prefix


def multiline(string: str):
    """
    Remove formatting lines and remove new lines from left and right
    :param string:
    :return:
    """
    string = string.rstrip()
    string = string.lstrip('\n')
    match = re.search('^ +', string)
    lines = string.split('\n')

    indent = 0
    if match is not None:
        indent = len(match.group())

    lines = list(map(lambda line: remove_prefix(' ' * indent, line), lines))
    return '\n'.join(lines)


import os
from typing import get_args, TypeVar
import environs

from telgine.utils.get_class_fields import get_class_fields

T = TypeVar('T')


def load_config(class_: T, path_to_dot_env: str, ignore_case: bool = False) -> T:
    """
    Load config from class to global scope
    """
    env = environs.Env()
    env.read_env(path_to_dot_env)
    class_fields = get_class_fields(class_)
    if ignore_case:
        for class_field in class_fields:
            class_field.name = class_field.name.upper()
    for class_field in class_fields:
        value_from_env = os.environ.get(class_field.name)
        attr_value = getattr(env, class_field.type.__name__)(
            class_field.name) if value_from_env is not None and value_from_env != '' else class_field.value
        if class_field.type.__name__ == 'list' and len(get_args(class_field.type)) > 0:
            _type = get_args(class_field.type)[0]
            attr_value = list(map(lambda item: _type(item), attr_value))
        setattr(class_, class_field.name, attr_value)
    return class_



from aiogram import types
from aiogram import Dispatcher
from aiogram.dispatcher.middlewares import BaseMiddleware
import asyncio


class AskDataMiddleware(BaseMiddleware):
    """
    Needless middleware for working `ask()` function.
    When `ask()` application is calling, function is subscribe to this middleware and wait for message from user.
    When a message is caught then middleware give the function this message and remove it from subscribers.
    """
    futures: dict = {}
    dp: Dispatcher

    def __init__(self, dp: Dispatcher):
        super(AskDataMiddleware, self).__init__()
        self.dp = dp

    async def on_pre_process_message(self, msg: types.Message, data: dict):
        chat_id = msg.chat.id
        if chat_id in self.futures:
            future = self.futures[chat_id]
            # Send message to ask_chat to handler1
            future.set_result(msg)
            self.unsubscribe(chat_id)

    def subscribe(self, chat_id: int, future: asyncio.Future):
        self.futures[chat_id] = future

    def unsubscribe(self, chat_id: int):
        del self.futures[chat_id]



import re


def remove_prefix(prefix: str, string: str) -> str:
    match = re.search(f'^{prefix}', string)
    if match is not None:
        return string[len(match.group()):]
    return string


def remove_postfix(postfix: str, string: str) -> str:
    match = re.search(f'{postfix}$', string)
    if match is not None:
        return string[:len(string) - len(match.group())]
    return string



import pathlib
import shutil
import os
import telgine


def initialize_project(project_path):
    templates_folder_path = pathlib.Path(__file__).parent.parent.joinpath('templates/project')
    generated_project_path = pathlib.Path(project_path).resolve()
    print(f'Initializing project {generated_project_path.name}')
    shutil.copytree(templates_folder_path, generated_project_path)


from sly import Parser
from .lexer import TelgineLexer
from .nodes import Tag, If, For, Else, Text
from dataclasses import dataclass
from .nodes import Expr

# --- Navigation Map ---
# Use search in your IDE by following labels:
# - TAG
# - ASSIGN VALUES
# - DIRECTIVES
# - ATTRIBUTES


@dataclass
class Directives:
    if_: any
    else_: any
    for_: any


@dataclass
class TagMeta:
    args: dict
    directives: Directives


class TelgineParser(Parser):
    tokens = TelgineLexer.tokens
    start = 'tag'

    @_('LT ID v_if attribute GT tags CLOSE ID GT')
    def v_if_branch(self, p):
        v_if: If = p[2]
        v_if.body = [Tag(p[1], p[5], p[3])]
        return v_if

    @_('LT ID v_else attribute GT tags CLOSE ID GT')
    def v_else_branch(self, p):
        return Tag(p[1], p[4], p[2])

    @_('v_if_branch v_else_branch')
    def tag(self, p):
        v_if: If = p[0]
        v_if.orelse = [p[1]]
        return v_if

    @_('v_if_branch')
    def tag(self, p):
        return p[0]

    @_('LT ID attribute GT tags CLOSE ID GT')
    def tag(self, p):
        return Tag(p[1], p[4], p[2])

    @_('TEXT')
    def tag(self, p):
        return Text(p[0])

    @_('tag')
    def tags(self, p):
        return [p.tag]

    @_('tags tags')
    def tags(self, p):
        return p[0] + p[1]

    @_('')
    def tags(self, p):
        return []

    @_('IF ASSIGN STRING')
    def v_if(self, p):
        return If(test=p.STRING)

    @_('ELSE ASSIGN STRING')
    def v_else(self, p):
        if len(p.STRING) > 0:
            raise SyntaxError('v-else statement must be have empty!')
        return Else()

    @_('ID ASSIGN STRING')
    def attribute(self, p):
        return {p.ID: p.STRING}

    @_('COLON ID ASSIGN STRING')
    def attribute(self, p):
        return {p.ID: Expr(p.STRING)}

    @_('attributes attributes')
    def attribute(self, p):
        return p[0] | p[1]

    @_('')
    def attributes(self, p):
        return {}








def parse(iterator):
    return TelgineParser().parse(iterator)[0]

# [token for token in tokens_gen]



from peewee import CharField, IntegrityError, Field, IntegerField, PrimaryKeyField, ForeignKeyField
from ..utils.get_class_fields import get_class_fields
from enum import Enum
from typing import Type


class Model:
    pass


class EnumField(CharField):
    def __init__(self, enumeration: Type[Enum], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.enumeration = enumeration

    def db_value(self, value: Enum) -> str:
        if not isinstance(value, self.enumeration):
            raise TypeError(f"Enum {self.enumeration.__name__} has no value '{value}'")
        return super().db_value(value.name)

    def python_value(self, value: str) -> Enum:
        try:
            return self.enumeration[super().python_value(value)]
        except KeyError:
            raise IntegrityError(
                f"Enum {self.enumeration.__name__} has no value with name '{value}'"
            ) from None


class RelationField(IntegerField):
    model_to: Type[Field]
    args: tuple
    kwargs: dict

    def __init__(self, model_to: Type[Field] | list[Type[Field]], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_to = model_to
        self.args = args
        self.kwargs = kwargs

    def db_value(self, value):
        pass

    def python_value(self, value):
        pass


def inverse_relations(models: list[Type[Field]]):
    for model in models:
        fields = get_class_fields(model)
        for field in fields:
            if hasattr(field.value, '__class__') and isinstance(field.value, RelationField):
                field_value: RelationField = field.value
                model_to = field_value.model_to

                setattr(model_to, model.__name__.lower(), ForeignKeyField(model, backref=getattr(model_to, '_meta').db_table, *field_value.args, **field_value.kwargs))
                delattr(model, field.name)




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['telgine',
 'telgine.cli',
 'telgine.core',
 'telgine.core.templates',
 'telgine.templates.project',
 'telgine.templates.project.bot',
 'telgine.templates.project.bot.config',
 'telgine.templates.project.bot.controllers',
 'telgine.templates.project.bot.filters',
 'telgine.templates.project.bot.middlewares',
 'telgine.templates.project.bot.models',
 'telgine.templates.project.bot.services',
 'telgine.templates.project.bot.stores',
 'telgine.templates.project.bot.templates',
 'telgine.templates.project.bot.types',
 'telgine.templates.project.bot.utils',
 'telgine.types',
 'telgine.utils']

package_data = \
{'': ['*'],
 'telgine': ['templates/*'],
 'telgine.templates.project.bot': ['locales/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'Werkzeug>=2.1.2,<3.0.0',
 'aiogram>=2.21,<3.0',
 'aioredis>=2.0.1',
 'aioschedule>=0.5.2,<0.6.0',
 'click>=8.1.3,<9.0.0',
 'cogapp>=3.3.0,<4.0.0',
 'environs>=9.5.0',
 'peewee>=3.14.10',
 'ply>=3.11,<4.0',
 'sly>=0.4,<0.5']

entry_points = \
{'console_scripts': ['tel = telgine.cli:main', 'telgine = telgine.cli:main']}

setup_kwargs = {
    'name': 'telgine',
    'version': '0.2.1',
    'description': 'Telegram bot framework',
    'long_description': '# telgine is the python telegram bot framework \n\n## Установка\nПоддерживаемая версия - `Python 3.10` и выше\n\nВажно! Если вы на Windows, то запускаете скрипт ниже из консоли запущенной с правами админимтратора, иначе скрипт не добавится в переменную `PATH` и команды `ni` и `telgine` работать не будут.\n\nНапишите `pip install telgine` для установки пакета из [PyPI](https://pypi.org).\n\nДля разработки этого пакета требуется установить [Poetry](https://python-poetry.org/) командой `pip install poetry`. Обязательно посмотрите [документацию](https://python-poetry.org/docs/) этого пакетного менеджера!\n\nНапишите `ni init superbot` для инициализации проекта.\n\n## Проект\n### Конфигурация\n`bot/config/env.py` Файл для описния переменных среды. \n\n```python\nclass EnvConfig:\n    BOT_NAME: str\n    BOT_TOKEN: str\n    ADMINS: list[int]\n\n    APP_HOST: str = \'localhost\'\n    APP_PORT: int = 8443\n    DB_HOST: str\n    DB_PORT: int\n    DB_NAME: str\n    DB_USER: str\n    DB_PASSWORD: str\n```\nВзгляните на этот кусок кода:\n```python\nAPPHOST: str = \'localhost\'\n```\n`APP_HOST` это поле класса в которое будет подргужено значение из файла `.env` по ключу `APP_HOST`. \n\n`str` означает, что при подргрузке переменной, ее значение будет приведено к этому типу. \n\n`\'localhost\'` - это значение по умолчанию, которое будет использовано в случае отсутствия ключа или его значения в файле `.env`\n## Контроллеры \n`bot/controllers` - папака для контроллеров\nВсе контроллеры подгружаются и регистрируются автоматически при инициализации приложения.\n\n```python\nfrom telgine import command, send\nfrom bot.config import config\n\n\n@command\nasync def start():\n  await send(f\'Application started at {config.APP_HOST}:{config.APP_PORT}\')\n```\n`@command` - это декоратор, который помечает обработчик команд `start()` как команду. Если вы отправите боту `/start`, то он исполнит тело этого обработчика.\nПо умочанию за имя команду принимается название функции. Вы также можете изменить его изменив значение параметра `name`.\nДля добавления описания функции, которое будет отображатся в списке команд Telegram, используйте поле `description`.\n```python\n@command(name=\'start\', description=\'Start this bot for you\')\ndef rename_me(): pass\n```\n\n`send()` - функция, которая отсылает текст пользователю бота, который вызвал обработку этого коллбека. \n### Получение данных от пользователя \nДля получения данных и текста введеного пользователем, вызовите функцию `message()`\n```python\nfrom telgine import command, message\n\n@command\nasync def start():\n    msg = message()\n    print(msg)\n```\n### Запрос данных у пользователя \n`ask()` Отсылает сообщение с просьбой ввести информацию и возвращает обьект сообщения с этой информацией.\n\n```python\nfrom telgine import command, ask, send\n\n\n@command\nasync def start():\n  name_msg = await ask(\'Введите ваше имя\')\n  age_msg = await ask(\'Теперь введите ваш возраст\')\n\n  print(name_msg)\n  print(age_msg)\n\n  send(f\'Привет {name_msg.text}, которому {age_msg.text} лет\')\n```\n### Другие декораторы для создания контроллеров\n`@startup` - Запускает обработчик при запуске приложения \n\n`@halt` - Запускает обработчик при остановке приложения\n\n`@hear` - Запускает обработчик при получение сообщения с обычным текстом текста\n\n\n## Хранение данных вне обработчиков\nДля сохранения данных введенных пользователем между вызовами разных обработчиков, нужно использовать `UserStore`\n```python\n# bot/stores/user.py\nfrom telgine import UserStore\n\nclass User(UserStore):\n    # \'Петр\' и 10 - это значение по умолчанию \n    name: str = \'Рома\'\n    age: int = 16\n```\n```python\n# bot/controllers/start.py\nfrom telgine import ask, command\nfrom bot.stores.User import User\n\n@command\nasync def start():\n    name = await ask(\'Введите ваше имя\')\n    age = await ask(\'Введите ваш возраст\')\n    user = User()\n    user.name = name.text, \n    user.age = int(age.text)\n\n@command\nasync def info():\n    user = User()\n    await send(f\'Привет {user.name}, которому {user.age} лет\')\n```\nЗдесь класс `User` - сохраняет данные не глобально для всех пользователей, а локально именно для этого пользователя в хранилеще по `id` его чата в `Storage`. По умолчанию используется `MemoryStorage`, но потом мы добавим `RedistStorage`, `MongoStorage` и `FileStorage`\n\n`UserStore.clear()` - востанавливает значение хранилища к значениям по умолчанию, если их нет, то удаляет все данные для этого пользователя.\n`UserStore.delete()` - полностью удаляет данные этого хранилища для этого пользователя.\n\n## Tasks\n- [ ] Implement modules autoimport \n- [ ] Implement method `select(message: str, enum: Enum)` which send a message with selection and return option of enum and map selected option to value\n- [ ] Implement `LocalStorage()` with the same syntax `name, email, age = LocalStore()` which create storage for each user. For implementation need to use python AST.\n- [ ] Make architecture for this project\n- [ ] Refactor all project\n- [ ] Make documentation better\n- [ ] Make errors better\n- [ ] Make using Config class from project for type auto complete\n- [ ] Add code formatter - black\n- [ ] Set name to "bot" folder the same as project\n- [ ] Make plugin system\n- [ ] Add Template renderer\n- [ ] Add `@startup`, `@stop` decorators\n- [ ] Add `@hear` decorator\n- [ ] Add the ability to work with inline query\n- [ ] Add checking for existence value in .env file. If value is not defined then raise error\n- [ ] Colorize output\n- [ ] Make beautiful serve message\n- [ ] Add the ability to adding other threads to app\n- [ ] Add `@middleware` decorator\n- [ ] Add `@filter` decorator\n- [ ] Add questions to init project command\n- [ ] Add auto db injection to peewee models\n- [ ] Change every event syntax\n- [ ] Implement overwrite default `/help` command text and reaction for it\n- [ ] Implement `/myuserid` command for improve user experience\n- [ ] Implement `setcommand()` function for showing commands at list\n- [ ] Schedulling with localization\n- [ ] Add logging with structuring for years, month and days\n- [ ] Buy website for telgine project\n- [ ] Make additional cli commands:\n  - [ ] `generate` - generate something project items\n  - [ ] `start` - start server\n  - [ ] `dev` - start server with auto reloading\n  - [ ] `build` - compile all project to one .pyc file\n  - [ ] `deploy` - deploy bot to telgine server\n  - [ ] `auth` - authorize for telgine server for deploy\n- [ ] Implement plugins:\n  - [ ] Role plugin\n  - [ ] Admin plugin\n  - [ ] Analytic plugin\n  - [ ] Payment plugin\n- [ ] Add notification about startup and stopping for bot admins\n- [ ] Detach telgine from specific service (Telegram) and make it universe:\n  - [ ] Telegram\n  - [ ] WhatsApp\n  - [ ] Discord\n  - [ ] Github\n  - [ ] Twitter and others social networks\n- [ ] Draw logotype for telgine project\n- [ ] Add choosing between requirements.txt and pyproject.toml files (poetry and pip options).',
    'author': 'Danil Sokolov',
    'author_email': 'danilzyx@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sejjax/telgine',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10',
}


setup(**setup_kwargs)



import pathlib
import os
import jinja2
import enum


def generate_artifacts(object_type, generated_file):
    def write_render_template(template_path: str, out_path: str) -> str:
        out_file_name = pathlib.Path(out_path).name
        out_path = pathlib.Path(out_path).parent.joinpath(out_file_name.capitalize())
        out_file_name = pathlib.Path(out_path).stem

        print(f'{template_path} {out_path}')
        env = jinja2.Environment(
            loader=jinja2.PackageLoader('telgine.py'),
            autoescape=jinja2.select_autoescape()
        )
        template = env.get_template(template_path)
        out = template.render({
            'name': out_file_name
        })
        with open(out_path, 'w') as file:
            file.write(out)

    class ObjectTypes(enum.Enum):
        MODEL = 'model'
        CONTROLLER = 'controller'
        FILTER = 'filter'
        STORE = 'store'
        TEMPLATE = 'template'

    def get_absolute_out_path(relative_out_path: str) -> str:
        """

        :param relative_out_path: path started at bot directory
        :return: {project_dir}/bot/{relative_out_path}
        """
        return str(pathlib.Path(os.getcwd()).joinpath(f'bot/{relative_out_path}').resolve())

    def write_render(relative_template_path: str, relative_out_path: str):
        write_render_template(
            relative_template_path,
            get_absolute_out_path(relative_out_path)
        )
    object_type: enum.Enum = ObjectTypes(object_type)

    if object_type == ObjectTypes.MODEL:
        write_render('model.txt', f'models/{generated_file}.py')
    elif object_type == ObjectTypes.STORE:
        write_render('store.txt', f'stores/{generated_file}.py')
    elif object_type == ObjectTypes.CONTROLLER:
        write_render('controller.txt', f'controllers/{generated_file}.py')
    elif object_type == ObjectTypes.FILTER:
        write_render('filter.txt', f'filters/{generated_file}.py')




import glob
import pathlib
import shutil
import hashlib


def check_tree_identity(path_a, path_b) -> bool:
    path_a = pathlib.Path(path_a)
    path_b = pathlib.Path(path_b)
    # If paths is not absolute then make it absolute
    if not path_a.is_absolute():
        path_a = path_a.resolve()
    if not path_b.is_absolute():
        path_b = path_b.resolve()

    # Get all nodes in subtrees
    items_a = glob.glob('**', root_dir=path_a, recursive=True)
    items_b = glob.glob('**', root_dir=path_b, recursive=True)
    # Compare number of nodes in subtree
    if len(items_a) != len(items_b):
        return False
    for item_a, item_b in zip(items_a, items_b):
        # Compare names of node
        if item_a != item_b:
            return False
        item_a = pathlib.Path(path_a).joinpath(item_a).resolve()
        item_b = pathlib.Path(path_b).joinpath(item_b).resolve()
        # If nodes is a dirrectory the skip data reading
        if item_a.is_dir() and item_b.is_dir():
            continue
        with open(item_a, 'rb') as f_a:
            data_a = f_a.read()
        with open(item_b, 'rb') as f_b:
            data_b = f_b.read()
        hash_a = hashlib.md5(data_a).digest()
        hash_b = hashlib.md5(data_b).digest()
        if hash_a != hash_b:
            return False
    return True
